import torch

from attacks import Attack
import torch.nn.functional as F

from constants import DEVICE
from utils import cross_entropy_loss, de_normalization, normalization


class SINIFGSM(Attack):
    """ SIM + NI-FGSM """

    def __init__(self, model, eps=16 / 255, steps=10, decay=1.0, m=5):
        """
        :param model: DNN model
        :param eps: the maximum perturbation
        :param steps: the number of iterations
        :param decay: the decay factor
        :param m: the number of scale copies
        """
        super().__init__("SINIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.alpha = self.eps / self.steps
        self.decay = decay
        self.m = m

    def forward(self, images, labels):
        targets = F.one_hot(labels.type(torch.int64), 1000).float().to(DEVICE)
        images_de_normalized = de_normalization(images)
        images_min = torch.clamp(images_de_normalized - self.eps, min=0.0, max=1.0)
        images_max = torch.clamp(images_de_normalized + self.eps, min=0.0, max=1.0)

        adv = images.clone()
        g = torch.zeros_like(images)
        for _ in range(self.steps):
            adv_de_normalized = de_normalization(adv)
            adv_nes_de_normalized = adv_de_normalized + self.decay * self.alpha * g
            adv_nes = normalization(adv_nes_de_normalized)

            grad = torch.zeros_like(images)
            for i in range(self.m):
                adv_nes_scaled = adv_nes / 2 ** i
                logits = self.model(adv_nes_scaled)
                loss = cross_entropy_loss(logits, targets)
                grad += torch.autograd.grad(loss, adv_nes)[0]
            grad /= self.m
            g = self.decay * g + grad / torch.mean(torch.abs(grad), dim=(1, 2, 3), keepdim=True)

            adv_de_normalized = de_normalization(adv)
            adv_de_normalized = torch.clamp(adv_de_normalized + self.alpha * torch.sign(g), min=images_min,
                                            max=images_max)
            adv = normalization(adv_de_normalized)

        return adv
